"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.emitPList = exports.emitJSON = exports.meta = exports.schema = void 0;
const promises_1 = require("fs/promises");
const onigasm_1 = require("onigasm");
const path_1 = require("path");
const plist_1 = __importDefault(require("plist"));
exports.schema = "https://raw.githubusercontent.com/martinring/tmlanguage/master/tmlanguage.json";
/**
 * Special scope that indicates a larger construct that doesn't get a single color.
 * Expanded to meta.<key>.<grammar name> during emit.
 */
exports.meta = Symbol("meta");
let initialized = false;
async function initialize() {
    if (!initialized) {
        const onigasmPath = require.resolve("onigasm");
        const wasmPath = (0, path_1.resolve)((0, path_1.dirname)(onigasmPath), "onigasm.wasm");
        const wasm = await (0, promises_1.readFile)(wasmPath);
        await (0, onigasm_1.loadWASM)(wasm.buffer);
        initialized = true;
    }
}
/**
 * Emit the given grammar to JSON.
 */
async function emitJSON(grammar, options = {}) {
    await initialize();
    const indent = 2;
    const processed = await processGrammar(grammar, options);
    return JSON.stringify(processed, undefined, indent);
}
exports.emitJSON = emitJSON;
/**
 * Emit the given grammar to PList XML
 */
async function emitPList(grammar, options = {}) {
    await initialize();
    const processed = await processGrammar(grammar, options);
    return plist_1.default.build(processed);
}
exports.emitPList = emitPList;
/**
 * Convert the grammar from our more convenient representation to the
 * tmlanguage schema. Perform some validation in the process.
 */
async function processGrammar(grammar, options) {
    await initialize();
    // key is rule.key, value is [unprocessed rule, processed rule]. unprocessed
    // rule is used for its identity to check for duplicates and deal with cycles.
    const repository = new Map();
    const output = processNode(grammar, options);
    output.repository = processRepository();
    return output;
    function processNode(node, options) {
        if (typeof node !== "object") {
            return node;
        }
        if (Array.isArray(node)) {
            return node.map((n) => processNode(n, options));
        }
        const output = {};
        for (const key in node) {
            const value = node[key];
            switch (key) {
                case "key":
                    // Drop it. It was used to place the node in the repository, and does
                    // not need to be retained on the node in the final structure.
                    break;
                case "scope":
                    // tmlanguage uses "name" confusingly for scope. We avoid "name" which
                    // can be confused with the repository key.
                    output.name = value === exports.meta ? `meta.${node.key}.${grammar.name.toLowerCase()}` : value;
                    break;
                case "begin":
                case "end":
                case "match":
                    validateRegexp(value, node, key, options);
                    output[key] = value;
                    break;
                case "patterns":
                    output[key] = processPatterns(value, options);
                    break;
                default:
                    output[key] = processNode(value, options);
                    break;
            }
        }
        return output;
    }
    function processPatterns(rules, options) {
        for (const rule of rules) {
            if (!repository.has(rule.key)) {
                // put placeholder first to prevent cycles
                const entry = [rule, undefined];
                repository.set(rule.key, entry);
                // fill placeholder with processed node.
                entry[1] = processNode(rule, options);
            }
            else if (repository.get(rule.key)[0] !== rule) {
                throw new Error("Duplicate key: " + rule.key);
            }
        }
        return rules.map((r) => ({ include: `#${r.key}` }));
    }
    function processRepository() {
        const output = {};
        for (const key of [...repository.keys()].sort()) {
            output[key] = repository.get(key)[1];
        }
        return output;
    }
    function validateRegexp(regexp, node, prop, options) {
        var _a;
        try {
            new onigasm_1.OnigRegExp(regexp).testSync("");
        }
        catch (err) {
            if (/^[0-9,]+$/.test(err.message)) {
                // Work around for https://github.com/NeekSandhu/onigasm/issues/26
                const array = new Uint8Array(err.message.split(",").map((s) => Number(s)));
                const buffer = Buffer.from(array);
                err = new Error(buffer.toString("utf-8"));
            }
            const sourceFile = (_a = options.errorSourceFilePath) !== null && _a !== void 0 ? _a : "unknown_file";
            //prettier-ignore
            // eslint-disable-next-line no-console
            console.error(`${sourceFile}(1,1): error TM0001: Bad regex: ${JSON.stringify({ [prop]: regexp })}: ${err.message}`);
            // eslint-disable-next-line no-console
            console.error(node);
            throw err;
        }
    }
}
//# sourceMappingURL=tmlanguage-generator.js.map